package eu.jonahbauer.wizard.client.libgdx.state;

import eu.jonahbauer.wizard.client.libgdx.Client;
import eu.jonahbauer.wizard.client.libgdx.screens.WaitingScreen;
import eu.jonahbauer.wizard.common.messages.client.LeaveSessionMessage;
import eu.jonahbauer.wizard.common.messages.client.ReadyMessage;
import eu.jonahbauer.wizard.common.messages.data.PlayerData;
import eu.jonahbauer.wizard.common.messages.data.SessionData;
import eu.jonahbauer.wizard.common.messages.server.*;
import eu.jonahbauer.wizard.common.model.Configuration;
import lombok.Getter;
import lombok.extern.log4j.Log4j2;

import java.util.*;
import java.util.stream.Collectors;

@Log4j2
@Getter
public final class Session extends BaseState {
    private WaitingScreen sessionScreen;
    
    private final UUID self;
    private final String secret;

    private final UUID session;
    private final String sessionName;
    private final Configuration configuration;
    private final List<PlayerData> players = new ArrayList<>();

    private boolean sending;

    public Session(SessionData session, Collection<PlayerData> players, UUID self, String secret) {
        this.session = session.getUuid();
        this.sessionName = session.getName();
        this.configuration = session.getConfiguration();
        this.players.addAll(players);

        this.self = self;
        this.secret = secret;
    }

    @Override
    public Optional<ClientState> onEnter(Client client) {
        sessionScreen = new WaitingScreen(client.getGame());
        client.getGame().setScreen(sessionScreen);
        sessionScreen.setPlayers(players.toArray(new PlayerData[0]));
        sessionScreen.setReady(false);
        sessionScreen.setPlayerName(getName());
        sessionScreen.setSession(session, sessionName, configuration);

        return super.onEnter(client);
    }

    @Override
    public Optional<ClientState> onMessage(Client client, ServerMessage message) {
        if (message instanceof PlayerJoinedMessage join) {
            sessionScreen.addPlayer(join.getPlayer());
            players.add(join.getPlayer());
            return Optional.empty();
        } else if (message instanceof PlayerLeftMessage leave) {
            sessionScreen.removePlayer(leave.getPlayer());
            players.remove(new PlayerData(leave.getPlayer(), null, false));
            return Optional.empty();
        } else if (message instanceof PlayerModifiedMessage modified) {
            var player = modified.getPlayer();

            sessionScreen.modifyPlayer(player);
            players.remove(player);
            players.add(player);

            if (self.equals(player.getUuid())) {
                sessionScreen.setReady(player.isReady());
            }

            return Optional.empty();
        } else if (message instanceof StartingGameMessage) {
            return Optional.of(new Game(
                    self, session, sessionName, secret,
                    players.stream().collect(Collectors.toMap(PlayerData::getUuid, PlayerData::getName))
            ));
        } else if (sending && message instanceof NackMessage nack) {
            // TODO display error
            log.error(nack.getMessage());
            sending = false;
            sessionScreen.setSending(false);
            return Optional.empty();
        } else if (sending && message instanceof AckMessage) {
            sending = false;
            sessionScreen.setSending(false);
            return Optional.empty();
        } else {
            return unexpectedMessage(client, message);
        }
    }

    public Optional<ClientState> setReady(Client client, boolean ready) {
        if (sending) {
            log.warn("Please slow down");
        } else {
            sending = true;
            sessionScreen.setSending(true);
            client.send(new ReadyMessage(ready));
        }
        return Optional.empty();
    }

    public Optional<ClientState> toggleReady(Client client) {
        return setReady(client, !isReady());
    }

    public Optional<ClientState> leave(Client client) {
        client.send(new LeaveSessionMessage());
        return Optional.of(new AwaitingJoinLobby());
    }

    private boolean isReady() {
        for (PlayerData player : players) {
            if (self.equals(player.getUuid())) {
                return player.isReady();
            }
        }
        throw new NoSuchElementException();
    }

    private String getName() {
        for (PlayerData player : players) {
            if (self.equals(player.getUuid())) {
                return player.getName();
            }
        }
        throw new NoSuchElementException();
    }
}
